import importlib.machinery
import sys


if __name__ == "__main__":
    file = sys.argv[-1]
    loader = importlib.machinery.SourceFileLoader("<py_compile>", file)
    source_bytes = loader.get_data(file)
    try:
        code = loader.source_to_code(source_bytes, file)
    except SyntaxError as e:
        print(":".join(("error", str(e.lineno), str(e.offset))), file=sys.stderr)
        print(e)
        sys.exit(1)



from collections import deque
import sys
from typing import *

from io import StringIO
from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtCore import Qt, pyqtSignal
import pexpect
from onlinejudge.dispatch import submission_from_url

from rujaion.custom_popup import CustomPopup
from rujaion.table_view import TableView
from rujaion.util import WriteObj


class Console(QtWidgets.QTextEdit):
    writeLnSignal = pyqtSignal(object)
    writeSignal = pyqtSignal(object)

    def __init__(self, parent=None):
        super(Console, self).__init__(parent)
        font = QtGui.QFont()
        font.setFamily("DejaVu Sans Mono")
        font.setFixedPitch(True)
        font.setPointSize(9)
        self.setFont(font)

        self._buffer = StringIO()
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.__contextMenu)
        self.writeLnSignal.connect(self.__write_by_line)
        self.writeSignal.connect(self.__write)
        self.evcxr_proc = None
        self.run_evcxr()

        self.command_history = deque(maxlen=10)
        # HACK
        self.color_words = ("\033[91m", "\033[94m", "\033[0m")
        self.history_pointer = 0
        self.popup = None

    def run_evcxr(self):
        if self.evcxr_proc is not None:
            self.evcxr_proc.terminate()

        try:
            self.evcxr_proc = pexpect.spawn("evcxr")
        except pexpect.exceptions.ExceptionPexpect as e:
            self.writeLnSignal.emit(e.value)
            return
        self.evcxr_proc.expect(">> ")
        self.writeLnSignal.emit(self.evcxr_proc.before.decode())
        self.display_prefix()

    def terminate_evcxr(self):
        if self.evcxr_proc is None:
            return
        self.evcxr_proc.terminate()
        self.evcxr_proc = None

    def display_prefix(self):
        tc = self.textCursor()
        tc.movePosition(QtGui.QTextCursor.EndOfLine)
        tc.insertText("\n>> ")
        tc.movePosition(QtGui.QTextCursor.EndOfBlock)
        tc.movePosition(QtGui.QTextCursor.EndOfLine)
        self.ensureCursorVisible()

    def __write(self, write_obj: WriteObj):
        msg = write_obj.msg
        mode = write_obj.mode
        self.moveCursor(QtGui.QTextCursor.End)
        if isinstance(msg, bytes):
            msg = msg.decode()
        font = QtGui.QFont()
        if mode == "":
            font.setBold(False)
            self.setTextColor(QtCore.Qt.black)
        elif mode == "success":
            font.setBold(True)
            self.setTextColor(QtCore.Qt.blue)
        elif mode == "error":
            font.setBold(True)
            self.setTextColor(QtCore.Qt.red)
        elif mode == "gdb":
            font.setBold(False)
            self.setTextColor(QtCore.Qt.black)
            msg = "(gdb) " + msg

        self.setFont(font)
        msg = msg + "\n"
        self.insertPlainText(msg)
        self._buffer.write(msg)

    def __write_by_line(self, msg: Union[str, bytes]):
        last_submission_url = None
        if isinstance(msg, bytes):
            msg = msg.decode()
        for line in msg.split("\n"):
            if line.startswith("[+]"):
                submit_result_prefix = "[+] success: result: "
                if line.startswith(submit_result_prefix):
                    last_submission_url = line[len(submit_result_prefix) :]
                self.__write(WriteObj(line, mode="success"))
            elif (
                line.startswith("[-]")
                or line.startswith("[!]")
                or line.startswith("[ERROR]")
            ):
                self.__write(WriteObj(line, mode="error"))
            else:
                self.__write(WriteObj(line))
        if last_submission_url is not None:
            submission = submission_from_url(last_submission_url)
            if submission is not None:
                self.popup = CustomPopup(None, last_submission_url)
                self.popup.show()

    def __getattr__(self, attr):
        return getattr(self._buffer, attr)

    def __contextMenu(self):
        menu = self.createStandardContextMenu()
        self._addCustomMenuItems(menu)
        menu.exec_(QtGui.QCursor.pos())

    def _addCustomMenuItems(self, menu: QtWidgets.QMenu):
        menu.addSeparator()
        menu.addAction(u"Clear", self.clear)
        menu.addAction(u"Visualize Table", self.visualize_table)

    def visualize_table(self):
        tc = self.textCursor()
        self.tv = TableView()
        self.tv.visualize_tables(tc.selectedText())
        self.tv.show()

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        tc = self.textCursor()
        if (
            event.key() == Qt.Key_Up
            and tc.blockNumber() == self.document().blockCount() - 1
        ):
            if self.history_pointer < len(self.command_history) - 1:
                self.history_pointer += 1
            else:
                return
            cur_idx = len(self.command_history) - self.history_pointer
            try:
                last_command = self.command_history[cur_idx]
            except IndexError:
                self.history_pointer = 0
                return
            tc.select(QtGui.QTextCursor.LineUnderCursor)
            tc.removeSelectedText()
            self.display_prefix()
            self.insertPlainText(last_command)
            return
        if (
            event.key() == Qt.Key_Down
            and tc.blockNumber() == self.document().blockCount() - 1
        ):
            if self.history_pointer >= 1:
                self.history_pointer -= 1
            cur_idx = len(self.command_history) - self.history_pointer
            try:
                last_command = self.command_history[cur_idx]
            except IndexError:
                self.history_pointer = 0
                return
            tc.select(QtGui.QTextCursor.LineUnderCursor)
            tc.removeSelectedText()
            self.display_prefix()
            self.insertPlainText(last_command)
            return
        self.history_pointer = 0

        if (
            event.key() == Qt.Key_Left
            and tc.blockNumber() == self.document().blockCount() - 1
            and tc.columnNumber() <= 3
            and self.evcxr_proc is not None
        ):
            return
        if event.key() in (
            Qt.Key_Left,
            Qt.Key_Right,
            Qt.Key_Up,
            Qt.Key_Down,
            Qt.Key_PageUp,
            Qt.Key_PageDown,
            Qt.Key_Home,
            Qt.Key_End,
        ):
            super().keyPressEvent(event)
            return

        if event.modifiers() == QtCore.Qt.ControlModifier:
            super().keyPressEvent(event)
            return

        # Editable final line only.
        if tc.blockNumber() != self.document().blockCount() - 1:
            return
        if event.key() == 16777220:
            if self.evcxr_proc is not None:
                print("execute")
                command = self.document().toPlainText().split("\n")[-1][3:] + "\n"
                self.evcxr_proc.send(command.encode())
                self.command_history.append(
                    self.document().toPlainText().split("\n")[-1][3:]
                )
                # self.evcxr_proc.send(b'println!("hello")\n')
                self.evcxr_proc.expect(">> ")

                line = "\n".join(self.evcxr_proc.before.decode().split("\r\n")[1:-1])
                for word in self.color_words:
                    line = line.replace(word, "")
                self.writeLnSignal.emit("\n" + line)

                self.display_prefix()

            return
        super().keyPressEvent(event)


class QMainWindow(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.console = Console()
        self.setCentralWidget(self.console)


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    main = QMainWindow()
    main.show()
    app.exec_()



# From https://www.binpress.com/building-text-editor-pyqt-3/

from PyQt5 import QtGui, QtWidgets
from PyQt5.QtCore import Qt

import re


class FindTextEdit(QtWidgets.QTextEdit):
    def __init__(self, parent, enter_action):
        self.enter_action = enter_action
        super().__init__(parent)

    def keyPressEvent(self, e: QtGui.QKeyEvent) -> None:
        if e.key() == 16777220:  # Enter
            self.enter_action()
            return
        if e.key() == Qt.Key_Tab:
            self.parent().toggle_focus()
            return
        super().keyPressEvent(e)


class Find(QtWidgets.QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent
        self.lastMatch = None
        self.lastStart = 0
        self.initUI()

    def initUI(self):

        # Button to search the document for something
        findButton = QtWidgets.QPushButton("Find", self)
        findButton.clicked.connect(self.find)

        # Button to replace the last finding
        replaceButton = QtWidgets.QPushButton("Find and Replace", self)
        replaceButton.clicked.connect(self.replace)

        # Button to remove all findings
        allButton = QtWidgets.QPushButton("Replace all", self)
        allButton.clicked.connect(self.replaceAll)

        # Normal mode - radio button
        self.normalRadio = QtWidgets.QRadioButton("Normal", self)
        self.normalRadio.toggled.connect(self.normalMode)

        # Regular Expression Mode - radio button
        self.regexRadio = QtWidgets.QRadioButton("RegEx", self)
        self.regexRadio.toggled.connect(self.regexMode)

        # The field into which to type the query
        self.findField = FindTextEdit(self, self.find)
        self.findField.resize(250, 50)
        self.findField.setFocus()

        # The field into which to type the text to replace the
        # queried text
        self.replaceField = FindTextEdit(self, self.replace)
        self.replaceField.resize(250, 50)

        optionsLabel = QtWidgets.QLabel("Options: ", self)

        # Case Sensitivity option
        self.caseSens = QtWidgets.QCheckBox("Case sensitive", self)

        # Whole Words option
        self.wholeWords = QtWidgets.QCheckBox("Whole words", self)

        # Layout the objects on the screen
        layout = QtWidgets.QGridLayout()

        layout.addWidget(self.findField, 1, 0, 1, 4)
        layout.addWidget(self.normalRadio, 2, 2)
        layout.addWidget(self.regexRadio, 2, 3)
        layout.addWidget(findButton, 2, 0, 1, 2)

        layout.addWidget(self.replaceField, 3, 0, 1, 4)
        layout.addWidget(replaceButton, 4, 0, 1, 2)
        layout.addWidget(allButton, 4, 2, 1, 2)

        # Add some spacing
        spacer = QtWidgets.QWidget(self)

        spacer.setFixedSize(0, 10)

        layout.addWidget(spacer, 5, 0)

        layout.addWidget(optionsLabel, 6, 0)
        layout.addWidget(self.caseSens, 6, 1)
        layout.addWidget(self.wholeWords, 6, 2)

        self.setGeometry(300, 300, 360, 250)
        self.setWindowTitle("Find and Replace")
        self.setLayout(layout)

        # By default the normal mode is activated
        self.normalRadio.setChecked(True)

    def toggle_focus(self):
        if self.findField.hasFocus():
            self.replaceField.setFocus()
        else:
            self.findField.setFocus()

    def find(self):
        # Grab the parent's text
        text = self.parent.toPlainText()

        # And the text to find
        query = self.findField.toPlainText()

        # If the 'Whole Words' checkbox is checked, we need to append
        # and prepend a non-alphanumeric character
        if self.wholeWords.isChecked():
            query = r"\W" + query + r"\W"

        # By default regexes are case sensitive but usually a search isn't
        # case sensitive by default, so we need to switch this around here
        flags = 0 if self.caseSens.isChecked() else re.I

        if self.normalRadio.isChecked():
            # Use normal string search to find the query from the
            # last starting position
            self.lastStart = text.find(query, self.lastStart + 1)

            # If the find() method didn't return -1 (not found)
            if self.lastStart >= 0:
                end = self.lastStart + len(query)
                self.moveCursor(self.lastStart, end)
            else:
                # Make the next search start from the begining again
                self.lastStart = 0
                # self.moveCursor(QtGui.QTextCursor.atStart())

        else:
            # Compile the pattern
            pattern = re.compile(query, flags)

            start = self.lastMatch.start() + 1 if self.lastMatch else 0

            # The actual search
            self.lastMatch = pattern.search(text, start)
            if self.lastMatch:
                start = self.lastMatch.start()
                end = self.lastMatch.end()

                # If 'Whole words' is checked, the selection would include the two
                # non-alphanumeric characters we included in the search, which need
                # to be removed before marking them.
                if self.wholeWords.isChecked():
                    start += 1
                    end -= 1
                self.moveCursor(start, end)

    def replace(self):
        self.find()

        # Grab the text cursor
        cursor = self.parent.textCursor()

        # Security
        if self.lastMatch and cursor.hasSelection():
            # We insert the new text, which will override the selected
            # text
            cursor.insertText(self.replaceField.toPlainText())

            # And set the new cursor
            self.parent.setTextCursor(cursor)

    def replaceAll(self):
        # Initial find() call so that lastMatch is
        # potentially not None anymore
        self.find()

        # Replace and find until find is None again
        while self.lastMatch:
            self.replace()
            self.find()

    def regexMode(self):

        # First uncheck the checkboxes
        self.caseSens.setChecked(False)
        self.wholeWords.setChecked(False)

        # Then disable them (gray them out)
        self.caseSens.setEnabled(False)
        self.wholeWords.setEnabled(False)

    def normalMode(self):

        # Enable checkboxes (un-gray them)
        self.caseSens.setEnabled(True)
        self.wholeWords.setEnabled(True)

    def moveCursor(self, start, end):

        # We retrieve the QTextCursor object from the parent's QTextEdit
        cursor = self.parent.textCursor()

        # Then we set the position to the beginning of the last match
        cursor.setPosition(start)

        # Next we move the Cursor by over the match and pass the KeepAnchor parameter
        # which will make the cursor select the the match's text
        cursor.movePosition(
            QtGui.QTextCursor.Right, QtGui.QTextCursor.KeepAnchor, end - start
        )

        # And finally we set this new cursor as the parent's
        self.parent.setTextCursor(cursor)



from PyQt5 import QtWidgets

from rujaion import util


class SubmitDialog(QtWidgets.QDialog):
    def __init__(self, *args, url: str, lang: str):
        super().__init__(*args)
        self.url_edit = util.StateLessTextEdit(url, self)
        self.lang_edit = util.StateLessTextEdit(lang, self)
        self.dialogs = (
            ("Submit...", None),
            ("URL", self.url_edit),
            ("Language", self.lang_edit),
        )
        self.submitter = util.Commander(self.parent().console)
        self.resize(500, 100)
        self.draw()

    def draw(self, *args):
        main_layout = QtWidgets.QVBoxLayout()
        for name, widget in self.dialogs:
            if not widget:
                l_widget = QtWidgets.QGroupBox(name)
                l_widget.setStyleSheet(
                    """
                QGroupBox {
                    color: white;
                    border: 1px solid gray;
                    border-radius: 9px;
                    margin-top: 0.5em;
                }
                QGroupBox::title {
                    color: white;
                    subcontrol-origin: margin;
                    left: 10px;
                    padding: 0 3px 0 3px;
                }
                """
                )
                l_widget.setFlat(False)
                section_layout = QtWidgets.QFormLayout()
                l_widget.setLayout(section_layout)
                main_layout.addWidget(l_widget)
            else:
                section_layout.addRow(name, widget)
        submit_button = QtWidgets.QPushButton("Submit")
        submit_button.clicked.connect(self.submit)
        main_layout.addWidget(submit_button)
        self.setLayout(main_layout)
        # self.adjustSize()

    def submit(self):
        for name, widget in self.dialogs:
            try:
                widget.commit()
            except AttributeError:
                pass
        cmd = (
            "oj",
            "s",
            "-l",
            self.lang_edit.text(),
            "-y",
            self.url_edit.text(),
            self.parent().editor.fname,
            "--no-open",
        )
        print(cmd)
        self.submitter.cmd = cmd
        self.submitter.start()
        self.close()



import codecs
from collections import defaultdict
import os
import subprocess
from typing import *

from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import Qt

from rujaion import completer
from rujaion.command import finder
from rujaion import syntax
from rujaion import util

"""
Fixing Some Bugs on a Sunday Evening
------------------------------------

Whose bugs these are I think I know,
But now he works at 3DO;
He will not see me working here
To fix his code and make it go.

The saner folk must think it queer
To trace without the source code near
After a launch and frozen mouse
The weirdest stack crawl of the year.

I give my nodding head a shake
To see if I can stay awake
The only other thing to do
Is find some more coffeine to take.

This bug is pretty hard to nip,
But I have other ones to fix,
And tons to go before we ship,
And tons to go before we ship.


Written by David A. Lyons <dlyons@apple.com>, January 1994
(with apologies to Robert Frost)

Hey, it's fiction.  Close to reality in spirit,
but does not refer to a specific project, bug, Sunday,
or brand of soft drink.
"""


class Editter(QtWidgets.QPlainTextEdit):
    toggleBreakSignal = pyqtSignal(bytes)

    def __init__(self, *args):
        super().__init__(*args)
        font = QtGui.QFont()
        font.setFamily("DejaVu Sans Mono")
        font.setFixedPitch(True)
        font.setPointSize(10)
        self.setFont(font)

        self.lineNumberAreaWidth = self.fontMetrics().width("8") * 4
        self.setViewportMargins(self.lineNumberAreaWidth, 0, 0, 0)
        self.lineNumberArea = QtWidgets.QWidget(self)
        self.lineNumberArea.installEventFilter(self)

        self.break_points = defaultdict(lambda: False)
        self.edited = False
        self.textChanged.connect(self.set_edited)
        self.fname = ""

        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.__contextMenu)
        self.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
        self.compile_error_selections = []

        self.lang = "rust"
        self.reset_lang()
        self.cursorPositionChanged.connect(self.highlight_cursor_line)
        self.first_error = None

    def __contextMenu(self):
        self._normalMenu = self.createStandardContextMenu()
        self._addCustomMenuItems(self._normalMenu)
        self._normalMenu.exec_(QtGui.QCursor.pos())

    def _addCustomMenuItems(self, menu: QtWidgets.QMenu):
        menu.addSeparator()
        menu.addAction(u"Go to declaration (Ctrl + b)", self.go_to_declaration)
        if (
            self.textCursor().selectedText()
            and "\u2029" not in self.textCursor().selectedText()
        ):
            menu.addAction(u"Add to watches", self.add_to_watch)

    def add_to_watch(self):
        var = self.textCursor().selectedText()
        self.parent().display_widget.add_var(var)

    def find(self):
        finder.Find(self).show()

    def default_file_name(self) -> str:
        if self.lang == "c++":
            return "test1.cpp"
        else:
            return "test1.rs"

    def highlight_cursor_line(self):
        if self.parent().debug_process is not None:
            return
        extra_selections = []
        selection = QtWidgets.QTextEdit.ExtraSelection()
        line_color = QtGui.QColor(Qt.yellow).lighter(180)

        selection.format.setBackground(line_color)
        selection.format.setProperty(
            QtGui.QTextFormat.FullWidthSelection, QtCore.QVariant(True)
        )
        selection.cursor = self.textCursor()
        # selection.cursor.clearSelection()
        extra_selections.append(selection)
        self.setExtraSelections(extra_selections + self.compile_error_selections)

    def highlight_compile_error(self, highlight_places, is_warning: bool):
        if is_warning:
            line_color = QtGui.QColor(Qt.yellow).darker(180)
        else:
            line_color = QtGui.QColor(Qt.red).lighter(110)
        for line, pos in highlight_places:
            self.add_highlight_error(line, line_color, pos)
        self.setExtraSelections(self.compile_error_selections)

    def reset_compile_info(self):
        self.compile_error_selections.clear()
        self.first_error = None

    def add_highlight_error(self, line: int, line_color: QtGui.QColor, pos: int):
        selection = QtWidgets.QTextEdit.ExtraSelection()
        selection.format.setFontUnderline(True)
        selection.format.setUnderlineColor(line_color)
        selection.format.setUnderlineStyle(QtGui.QTextCharFormat.WaveUnderline)
        selection.format.setProperty(
            QtGui.QTextFormat.FullWidthSelection, QtCore.QVariant(True)
        )
        cursor = QtGui.QTextCursor(self.document().findBlockByLineNumber(line - 1))
        cursor.movePosition(QtGui.QTextCursor.StartOfLine)
        cursor.movePosition(
            QtGui.QTextCursor.Right, QtGui.QTextCursor.MoveAnchor, pos - 1
        )
        # self.setTextCursor(cursor)
        cursor.select(QtGui.QTextCursor.WordUnderCursor)
        selection.cursor = cursor
        self.compile_error_selections.append(selection)
        self.first_error = (line, pos)

    def eventFilter(self, obj, event: QtGui.QKeyEvent) -> bool:
        if obj is self.lineNumberArea and event.type() == QtCore.QEvent.Paint:
            self.drawLineNumbers()
            return True
        return False

    def drawLineNumbers(self):
        painter = QtGui.QPainter(self.lineNumberArea)
        painter.setPen(Qt.black)
        r = QtCore.QRect(self.lineNumberArea.rect())
        painter.fillRect(r, Qt.lightGray)
        ht = self.fontMetrics().height()
        block = QtGui.QTextBlock(self.firstVisibleBlock())
        line_number = block.blockNumber() + 1
        while block.isValid():
            y = self.blockBoundingGeometry(block).translated(self.contentOffset()).top()
            if y >= r.bottom():
                break
            if self.break_points[line_number]:
                disp = "b " + str(line_number)
            else:
                disp = str(line_number)
            painter.drawText(0, y, self.lineNumberAreaWidth, ht, Qt.AlignRight, disp)
            line_number += 1
            block = block.next()

    def resizeEvent(self, event: QtGui.QKeyEvent):
        super().resizeEvent(event)
        self.lineNumberArea.setGeometry(
            QtCore.QRect(
                self.rect().left(),
                self.rect().top(),
                self.lineNumberAreaWidth,
                self.rect().height(),
            )
        )

    def wheelEvent(self, event: QtGui.QKeyEvent):
        super().wheelEvent(event)
        self.repaint()

    def toggleBreak(self, line_num: int):
        self.break_points[line_num] = not self.break_points[line_num]
        self.repaint()
        if self.break_points[line_num]:
            command = ("b " + str(line_num) + "\n").encode()
        else:
            command = (str(line_num)).encode()
        self.toggleBreakSignal.emit(command)

    def generateBreak(self):
        commands = []
        for i in range(1, self.document().blockCount()):
            if self.break_points[i]:
                commands.append(("b " + str(i) + "\n").encode())
        return commands

    def highlight_executing_line(self, line_num: int):
        extraSelections = []
        selection = QtWidgets.QTextEdit.ExtraSelection()
        selection.format.setBackground(Qt.cyan)
        selection.format.setProperty(
            QtGui.QTextFormat.FullWidthSelection, QtCore.QVariant(True)
        )
        selection.cursor = QtGui.QTextCursor(
            self.document().findBlockByLineNumber(line_num - 1)
        )
        selection.cursor.clearSelection()
        extraSelections.append(selection)
        self.setExtraSelections(extraSelections)

        self.setTextCursor(selection.cursor)
        self.ensureCursorVisible()

    def clear_highlight_line(self):
        self.setExtraSelections([])

    def open_file(self, fname: str):
        assert fname
        with open(fname) as f:
            self.setPlainText(f.read())
        self.edited = False
        self.fname = fname
        self.reset_lang()

    def reset_file_name(self):
        if self.fname:
            default_dir = os.path.dirname(self.fname)
        else:
            default_dir = os.path.dirname(__file__) + "/rust_sample"
        self.fname = os.path.join(default_dir, self.default_file_name())

    def reset_lang(self, fname: Optional[str] = None):
        if fname is None:
            fname = self.fname
        if fname.endswith("cpp") or fname.endswith("cc"):
            self.lang = "c++"
        elif fname.endswith("py"):
            self.lang = "python3"
        else:
            self.lang = "rust"

        if self.lang == "rust":
            self.completer = completer.RacerCompleter(self)
            self.highlighter = syntax.RustHighlighter(self.document())
        elif self.lang == "python3":
            self.completer = completer.PyCompleter(self)
            self.highlighter = syntax.PyHighlighter(self.document())
        else:
            self.completer = completer.CppCompleter(self)
            self.highlighter = syntax.CppHighlighter(self.document())
        self.completer.setWidget(self)
        self.parent().lang_box.setCurrentText(self.lang)

    def new_file(self, template_file_name: str):
        self.break_points = defaultdict(lambda: False)
        self.clear()
        self.edited = False
        self.reset_lang(template_file_name)
        self.reset_file_name()
        with open(template_file_name) as f:
            self.setPlainText(f.read())
        self.repaint()

    def set_edited(self):
        self.edited = True

    def go_to_declaration(self):
        if self.lang != "rust":
            # Not supported
            return
        src_line_num = str(self.textCursor().blockNumber() + 1)
        src_char_num = str(self.textCursor().columnNumber())

        try:
            # out = subprocess.check_output(("racer", "find-definition",
            #                                src_line_num, src_char_num,
            #                                self.fname))
            out = subprocess.check_output(
                "racer find-definition"
                + " "
                + src_line_num
                + " "
                + src_char_num
                + " "
                + self.fname,
                shell=True,
            ).decode()
        except Exception:
            return
        if not out.startswith("MATCH"):
            return
        out = out[6:]
        words = out.split(",")
        line_num, char_num = int(words[1]), int(words[2])
        self.go_to(line_num, char_num)

    def go_to(self, line_num, char_num):
        cursor = QtGui.QTextCursor(self.document().findBlockByLineNumber(line_num - 1))
        cursor.movePosition(
            QtGui.QTextCursor.NextCharacter, QtGui.QTextCursor.MoveAnchor, char_num
        )
        self.setTextCursor(cursor)
        self.ensureCursorVisible()

    def go_to_first_error(self):
        if self.first_error is None:
            return
        line_num, char_num = self.first_error
        self.go_to(line_num, char_num)

    def insertCompletion(self):
        text = self.completer.getSelected()
        if text not in self.completer.candidates_dict:  # Not selected from pop up
            self.completer.popup().hide()
            return

        tc = self.textCursor()

        # First we should remove char since some live templates'
        # completion does not start with completion prefix.
        for i in range(len(self.completer.completionPrefix())):
            tc.movePosition(QtGui.QTextCursor.Left)
            tc.deleteChar()

        # tc.movePosition(QtGui.QTextCursor.EndOfWord)
        tc.insertText(text)
        if self.completer.candidates_dict[text] != -1:
            for _ in range(self.completer.candidates_dict[text]):
                tc.movePosition(QtGui.QTextCursor.Left)
        self.setTextCursor(tc)
        self.completer.popup().hide()

    def enter_with_auto_indent(self):
        line_number = self.textCursor().blockNumber()
        line_text = self.document().findBlockByLineNumber(line_number).text()
        indent_level = line_text.count(" " * util.indent_width(self.lang))
        if self.lang == "python3":
            if line_text.endswith(":"):
                indent_level += 1
        elif line_text.endswith("{"):
            indent_level += 1
        self.insertPlainText("\n" + " " * util.indent_width(self.lang) * indent_level)

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        tc = self.textCursor()
        if event.key() == 16777220 and self.completer.popup().isVisible():
            self.insertCompletion()
            self.completer.setCompletionMode(QtWidgets.QCompleter.PopupCompletion)
            return

        if event.key() == 16777220:  # Enter
            self.enter_with_auto_indent()
            return

        if event.key() == Qt.Key_BraceRight:
            # Decrease indent level
            if (
                tc.position() > 0
                and self.document().characterAt(tc.position() - 1) != "{"
            ):
                tc.movePosition(QtGui.QTextCursor.StartOfLine)
                for i in range(4):
                    if self.document().characterAt(tc.position()) == " ":
                        tc.deleteChar()

        if event.key() == Qt.Key_F5:
            line_num = tc.blockNumber() + 1
            self.toggleBreak(line_num)
            return

        # comment out or uncomment
        if (
            event.modifiers() == QtCore.Qt.ControlModifier
            and event.key() == Qt.Key_Slash
        ):
            tc = self.textCursor()
            if tc.selection().isEmpty():
                selected_line_num = 1
            else:
                selected_line_num = tc.selection().toPlainText().count("\n") + 1
            for i in range(selected_line_num):
                tc.movePosition(QtGui.QTextCursor.StartOfLine)
                if self.document().characterAt(tc.position()) == " ":
                    tc.movePosition(
                        QtGui.QTextCursor.NextWord, QtGui.QTextCursor.MoveAnchor, 1
                    )
                if (
                    self.document().characterAt(tc.position()) == "/"
                    and self.document().characterAt(tc.position() + 1) == "/"
                    and self.document().characterAt(tc.position() + 2) == " "
                ):
                    tc.deleteChar()
                    tc.deleteChar()
                    tc.deleteChar()
                else:
                    tc.insertText("// ")
                tc.movePosition(QtGui.QTextCursor.Down, QtGui.QTextCursor.MoveAnchor, 1)
            return

        if event.key() == Qt.Key_Tab and event.modifiers() == QtCore.Qt.ControlModifier:
            tc = self.textCursor()
            if tc.selection().isEmpty():
                selected_line_num = 1
            else:
                selected_line_num = tc.selection().toPlainText().count("\n") + 1
            for i in range(selected_line_num):
                tc.movePosition(QtGui.QTextCursor.StartOfLine)
                if self.document().characterAt(tc.position()) == " ":
                    tc.deleteChar()
                    tc.deleteChar()
                    tc.deleteChar()
                    tc.deleteChar()
                    tc.movePosition(
                        QtGui.QTextCursor.Down, QtGui.QTextCursor.MoveAnchor, 1
                    )
            return

        if event.key() == Qt.Key_Tab:
            tc = self.textCursor()
            if tc.selection().isEmpty():
                selected_line_num = 1
            else:
                selected_line_num = tc.selection().toPlainText().count("\n") + 1
            for i in range(selected_line_num):
                tc.movePosition(QtGui.QTextCursor.StartOfLine)
                tc.insertText(" " * util.indent_width(self.lang))
                tc.movePosition(QtGui.QTextCursor.Down, QtGui.QTextCursor.MoveAnchor, 1)
            return

        if event.key() == Qt.Key_K and event.modifiers() == QtCore.Qt.ControlModifier:
            self.remove_whole_line()
            return
        if (
            event.modifiers() == QtCore.Qt.ControlModifier
            and event.key() == QtCore.Qt.Key_B
        ):
            self.go_to_declaration()
            return
        if event.key() == QtCore.Qt.Key_F2:
            self.go_to_first_error()
            return

        super().keyPressEvent(event)

        if event.key() == Qt.Key_Home:
            if self.textCursor().selectedText():
                return
            if self.document().characterAt(self.textCursor().position()) == " ":
                cursor = self.textCursor()
                cursor.movePosition(
                    QtGui.QTextCursor.NextWord, QtGui.QTextCursor.MoveAnchor, 1
                )
                self.setTextCursor(cursor)

        self.start_complete_process(event, tc)

    # TODO here is slow.
    def start_complete_process(self, event: QtGui.QKeyEvent, tc: QtGui.QTextCursor):
        if (
            event.text().isalnum()
            and not self.document().characterAt(tc.position() + 1).isalnum()
        ):
            tc.movePosition(QtGui.QTextCursor.Left, QtGui.QTextCursor.MoveAnchor, 1)
            tc.select(QtGui.QTextCursor.WordUnderCursor)
            if len(tc.selectedText()) <= 1:
                return
            self.completer.setCompletionPrefix(tc.selectedText())
            popup = self.completer.popup()
            popup.setCurrentIndex(self.completer.completionModel().index(0, 0))
            cr = self.cursorRect()
            cr.setWidth(
                self.completer.popup().sizeHintForColumn(0)
                + self.completer.popup().verticalScrollBar().sizeHint().width()
            )
            self.completer.complete(cr)
        else:
            self.completer.popup().hide()

    def remove_whole_line(self):
        cursor = self.textCursor()
        cursor.select(QtGui.QTextCursor.LineUnderCursor)
        cursor.removeSelectedText()
        cursor.deletePreviousChar()

    def save_pre_process(self):
        # Get formatted Text
        temp_file = codecs.open(util.TEMPFILE, "w", "utf-8")
        temp_file.write(self.toPlainText())
        temp_file.close()
        if not util.exec_format(self.lang):
            return
        temp_file = codecs.open(util.TEMPFILE, "r", "utf-8")
        all_text = "".join([line for line in temp_file.readlines()])
        temp_file.close()

        # Save cursor and scroll bar status
        cursor = self.textCursor()
        line_num = cursor.blockNumber()
        char_num = cursor.columnNumber()
        scroll_value = self.verticalScrollBar().value()

        # Clear all Text and insert format result
        cursor.movePosition(QtGui.QTextCursor.Start)
        cursor.movePosition(QtGui.QTextCursor.End, QtGui.QTextCursor.KeepAnchor, 1)
        cursor.removeSelectedText()
        self.insertPlainText(all_text)

        # recover cursor and scroll bar status
        cursor = QtGui.QTextCursor(self.document().findBlockByLineNumber(line_num))
        cursor.movePosition(
            QtGui.QTextCursor.NextCharacter, QtGui.QTextCursor.MoveAnchor, char_num
        )
        self.setTextCursor(cursor)
        self.verticalScrollBar().setValue(scroll_value)
        self.edited = False
        self.repaint()



import sys

import pandas

from PyQt5 import QtWidgets
from PyQt5.QtWebEngineWidgets import QWebEngineView, QWebEngineScript
from PyQt5.QtCore import Qt


table_style = (
    "table{width: 50%; border-collapse: collapse; border-spacing: 0;}"
    "table th,table td{ padding: 3px 0; text-align: center;} "
    "table tr:nth-child(odd){background-color: #ddd} "
    "body{font-family: arial}"
)


class TableView(QWebEngineView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.insertStyleSheet()
        self.setWindowFlags(Qt.WindowStaysOnTopHint)
        self.setMaximumSize(700, 600)
        self.setWindowTitle("Table View")

    def insertStyleSheet(self):
        name = "table_style"
        s = """(function() {
    css = document.createElement('style');
    css.type = 'text/css';
    css.id = '%s';
    document.head.appendChild(css);
    css.innerText = '%s';
})()""" % (
            name,
            table_style,
        )
        self.page().runJavaScript(s, QWebEngineScript.ApplicationWorld)
        script = QWebEngineScript()
        script.setName(name)
        script.setSourceCode(s)
        script.setInjectionPoint(QWebEngineScript.DocumentReady)
        script.setRunsOnSubFrames(True)
        script.setWorldId(QWebEngineScript.ApplicationWorld)
        self.page().scripts().insert(script)

    def visualize_tables(self, text: str):
        htmls = []
        for line in text.split(u"\u2029"):
            if "[" in line:
                start_idx = line.find("[")
                end_idx = line.rfind("]")
                try:
                    table = pandas.read_json(line[start_idx : end_idx + 1])
                except ValueError:
                    continue
                htmls.append(table.to_html())
                print(table.to_html())
        html = "<br>".join(htmls)
        self.setHtml(html)


if __name__ == "__main__":
    text = (
        "v = [[0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0, 0, 0]]\n"
        + "v = [[0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0, 0, 0]]"
    )
    app = QtWidgets.QApplication(sys.argv)
    ex = TableView()
    ex.visualize_tables(text)
    ex.show()
    sys.exit(app.exec_())



from PyQt5 import QtCore, Qt, QtGui
from PyQt5 import QtWidgets


class ResultTableModel(QtWidgets.QTableWidget):
    def __init__(self, parent=None, *args):
        super().__init__(parent, *args)
        self.resize(350, 800)
        self.row_size = 15
        self.column_size = 3
        self.setRowCount(self.row_size)
        self.setColumnCount(self.column_size)
        self.setHorizontalHeaderLabels(["Name", "Value", "Type"])
        self.setVerticalHeaderLabels([""] * self.row_size)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.__contextMenu)

        for i in range(self.row_size):
            for j in range(1, self.column_size):
                item = self.cell()
                # execute the line below to every item you need locked
                item.setFlags(QtCore.Qt.ItemIsEnabled)
                self.setItem(i, j, item)

    def __contextMenu(self):
        menu = QtWidgets.QMenu()
        self._addCustomMenuItems(menu)
        menu.exec_(QtGui.QCursor.pos())

    def _addCustomMenuItems(self, menu: QtWidgets.QMenu):
        menu.addSeparator()
        menu.addAction(u"Clear", self.clear)

    def cell(self, var: str = "") -> Qt.QTableWidgetItem:
        item = Qt.QTableWidgetItem()
        item.setText(var.replace("true", "T").replace("false", "F"))
        return item

    def set_cell(self, i: int, j: int, var: str = ""):
        self.setItem(i, j, self.cell(var))

    def name_iter(self):
        for i in range(self.row_size):
            if self.item(i, 0) is not None:
                yield i, self.item(i, 0).text()

    def clear(self):
        for i in range(self.row_size):
            self.set_cell(i, 0, var="")

    def add_var(self, var: str):
        for i in range(self.row_size):
            if self.item(i, 0) is None:
                self.set_cell(i, 0, var)
                return

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        pass



#!/usr/bin/env python3
import imp
from setuptools import find_packages

from setuptools import setup


def load_module(module_path):
    path = None
    for name in module_path.split("."):
        file, path, description = imp.find_module(name, path)
        path = [path]
    return imp.load_module(name, file, path[0], description)


setup(
    name="Rujaion",
    version="0.9.0",
    author="Ryosuke Fukatani",
    install_requires=[
        "PyQt5",
        "PyQtWebEngine",
        "pexpect",
        "online-judge-tools >= 7.2.2",
        "pandas",
    ],
    packages=find_packages(exclude=("docker", "doc")),
    package_data={"rujaion": ["resources/*"]},
    entry_points={"console_scripts": ["rujaion = rujaion.rujaion_main:main"]},
)



import codecs
from _collections import OrderedDict
import os
import subprocess
from typing import *
from xml.etree import ElementTree

from PyQt5 import QtWidgets, QtCore

from rujaion import util


class CompleterBase(QtWidgets.QCompleter):
    def __init__(self, parent=None):
        super().__init__((), parent)
        self.setCompletionMode(QtWidgets.QCompleter.UnfilteredPopupCompletion)
        self.highlighted.connect(self.setHighlighted)
        self.parent = parent
        self.live_templates = load_template(self.live_template_file)
        self.candidates_dict = {}
        self.ng_words = ["core", "proc_macro", "profiler_builtins"]
        self.lastSelected = ""

    def setHighlighted(self, text: str):
        self.lastSelected = text

    def getSelected(self) -> str:
        return self.lastSelected


class RacerCompleter(CompleterBase):
    # temp file for racer input
    live_template_file = os.path.join(
        util.get_resources_dir(), "live_templates_rust.xml"
    )

    # this is heavy?
    def setCompletionPrefix(self, text: str):
        temp_file = codecs.open(util.TEMPFILE, "w", "utf-8")
        temp_file.write(self.parent.toPlainText())
        temp_file.close()
        src_line_num = str(self.parent.textCursor().blockNumber() + 1)
        src_char_num = str(self.parent.textCursor().columnNumber())

        try:
            out = subprocess.check_output(
                "racer complete"
                + " "
                + src_line_num
                + " "
                + src_char_num
                + " "
                + util.TEMPFILE,
                shell=True,
            ).decode()
        except subprocess.CalledProcessError as e:
            print(e.output)
            return

        self.candidates_dict = {}
        for line in out.split("\n"):
            if line.startswith("MATCH"):
                cand = line[6:].split(",")[0]
                if cand not in self.ng_words:
                    self.candidates_dict[cand] = -1
        search_word = out.split("\n")[0].split(",")[2]

        for live_template in self.live_templates:
            if search_word in live_template.name:
                self.candidates_dict[live_template.template] = live_template.rpos
        if len(self.candidates_dict) >= 6 or search_word in self.candidates_dict.keys():
            self.candidates_dict = {}
        self.setModel(QtCore.QStringListModel(self.candidates_dict.keys()))
        super().setCompletionPrefix(search_word)


class PyCompleter(CompleterBase):
    live_template_file = os.path.join(
        util.get_resources_dir(), "live_templates_cpp.xml"
    )

    def setCompletionPrefix(self, text: str):
        import jedi

        src_line_num = self.parent.textCursor().blockNumber() + 1
        src_char_num = self.parent.textCursor().columnNumber()
        candidates = jedi.Script(
            self.parent.toPlainText(), src_line_num, src_char_num
        ).completions()
        candidates = [cand.name for cand in candidates]

        self.candidates_dict = {}
        for i, word in enumerate(candidates):
            self.candidates_dict[word] = -1

        self.setModel(QtCore.QStringListModel(self.candidates_dict.keys()))
        super().setCompletionPrefix(text)


class CppCompleter(CompleterBase):
    # temp file for racer input
    live_template_file = os.path.join(
        util.get_resources_dir(), "live_templates_cpp.xml"
    )

    # TODO: Support clang completer
    def setCompletionPrefix(self, text: str):
        temp_file = codecs.open(util.TEMPFILE_CPP, "w", "utf-8")
        temp_file.write(self.parent.toPlainText())
        temp_file.close()
        src_line_num = str(self.parent.textCursor().blockNumber() + 1)
        src_char_num = str(self.parent.textCursor().columnNumber())

        try:
            out = subprocess.check_output(
                # read all header is too slow
                # "clang -fsyntax-only -Xclang -code-completion-at=%s:%s:%s %s"
                "clang -cc1 -fsyntax-only -code-completion-at=%s:%s:%s %s"
                % (util.TEMPFILE_CPP, src_line_num, src_char_num, util.TEMPFILE_CPP),
                shell=True,
            ).decode()
        except subprocess.CalledProcessError as e:
            out = e.output.decode()

        self.candidates_dict = OrderedDict()
        for line in out.split("\n"):
            if line.startswith("COMPLETION:"):
                cand = line.split(" ")[1]
                if text not in cand:
                    continue
                if cand not in self.ng_words:
                    self.candidates_dict[cand] = -1

        for live_template in self.live_templates:
            if text in live_template.name:
                self.candidates_dict[live_template.template] = live_template.rpos
        if len(self.candidates_dict) >= 10 or text in self.candidates_dict.keys():
            self.candidates_dict = {}
        self.setModel(QtCore.QStringListModel(self.candidates_dict.keys()))
        super().setCompletionPrefix(text)


class LiveTemplate:
    def __init__(self):
        self.name = ""
        self.body = ""
        self.variables = []
        self.default_values = []
        self.template = ""
        self.rpos = -1

    def generate(self):
        self.template = self.body
        for var, value in zip(self.variables, self.default_values):
            if value is not None:
                self.template = self.template.replace("$" + var + "$", value)
        if "$END$" in self.template:
            self.rpos = len(self.template) - self.template.find("$END$") - 5
        self.template = self.template.replace("$END$", "")


def load_template(xml_file_name: str) -> List[LiveTemplate]:
    if not os.path.isfile(xml_file_name):
        return []
    with open(xml_file_name) as f:
        xml = f.read()
    root = ElementTree.fromstring(xml)
    templates = []
    for template in root.findall("template"):
        new_template = LiveTemplate()
        new_template.name = template.get("name")
        new_template.body = template.get("value")
        for variable in template.findall("variable"):
            new_template.variables.append(variable.get("name"))
            if "defaultValue" in variable.keys():
                new_template.default_values.append(variable.get("defaultValue"))
            else:
                new_template.default_values.append(None)
        new_template.generate()
        templates.append(new_template)
    return templates



import subprocess

from PyQt5 import QtWidgets

from rujaion import util


class LoginDialog(QtWidgets.QDialog):
    def __init__(self, *args, url: str, settings=None):
        super().__init__(*args)
        self.settings = settings
        self.password_edit = util.StateLessTextEdit("", self)
        self.password_edit.setEchoMode(QtWidgets.QLineEdit.Password)
        self.url_edit = util.StateLessTextEdit(url, self)
        self.dialogs = (
            ("Login Settings", None),
            ("URL", self.url_edit),
            ("Account Name", util.StateFullTextEdit(self.settings, "Account", self)),
            ("Password", self.password_edit),
        )
        self.resize(500, 100)
        self.draw()

    def draw(self, *args, settings=None):
        main_layout = QtWidgets.QVBoxLayout()
        for name, widget in self.dialogs:
            if not widget:
                l_widget = QtWidgets.QGroupBox(name)
                l_widget.setStyleSheet(
                    """
                QGroupBox {
                    color: white;
                    border: 1px solid gray;
                    border-radius: 9px;
                    margin-top: 0.5em;
                }
                QGroupBox::title {
                    color: white;
                    subcontrol-origin: margin;
                    left: 10px;
                    padding: 0 3px 0 3px;
                }
                """
                )
                l_widget.setFlat(False)
                section_layout = QtWidgets.QFormLayout()
                l_widget.setLayout(section_layout)
                main_layout.addWidget(l_widget)
            else:
                section_layout.addRow(name, widget)
        login_button = QtWidgets.QPushButton("Login")
        login_button.clicked.connect(self.login)
        main_layout.addWidget(login_button)
        self.setLayout(main_layout)

    def login(self):
        for name, widget in self.dialogs:
            try:
                widget.commit()
            except AttributeError:
                pass
        self.settings.sync()
        cmd = (
            "oj",
            "l",
            "-u",
            self.settings.value("Account", type=str),
            "-p",
            self.password_edit.text(),
            self.url_edit.text(),
        )
        try:
            out = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            self.parent().console.writeLnSignal.emit(out)
        except subprocess.CalledProcessError as err:
            self.parent().console.writeLnSignal.emit(err.output)
        self.close()


class URLEdit(QtWidgets.QLineEdit):
    def __init__(self, url: str, parent):
        super().__init__()
        self.parent = parent
        self.setText(url)

    def commit(self):
        pass
